#! /usr/bin/env python
"""
radec.py
"""

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon
import os
import numpy as np


# statically stored to save computation
# calculated with +- 10 degrees from poles
# can be re-generated with `generate_pole_field_ids`
POLE_FIELD_IDS = set([1,   33,   55,   56,   84,   85,  119,  120,  156,
                      157,  198, 199,  200,  244,  245,  246,  293,  294,
                      295,  344,  345,  394, 395,  396,  397,  446,  447,
                      497,  498,  499,  500,  549,  550, 551,  599,  600,
                      601,  647,  648,  649,  693,  694,  735,  736, 772,
                      773,  805,  806,  832,  833,  853,  878,  880, 1015,
                      1041, 1061, 1062, 1087, 1088, 1119, 1154, 1155, 1156,
                      1195, 1196, 1238, 1239, 1240, 1286, 1287, 1335, 1336,
                      1337, 1387, 1388, 1389, 1390, 1440, 1441, 1491, 1492,
                      1493, 1494, 1544, 1545, 1546, 1595, 1596, 1597, 1644,
                      1645, 1646, 1691, 1692, 1734, 1735, 1773, 1774, 1808,
                      1809, 1838, 1839, 1862, 1880, 1881])


def generate_pole_field_ids():
    ZTF_fields = load_ZTF_fields()
    ra = ZTF_fields['ra']
    cond = (ra <= 10) + (ra >= (360 - 10))
    return ZTF_fields['id'][cond]


def field_is_pole(field_id):
    return field_id in POLE_FIELD_IDS


def lightcurve_file_is_pole(lightcurve_file):
    field = int(lightcurve_file.split('_')[0].replace('field', ''))
    return field_is_pole(field)


def load_ZTF_CCD_corners(field_id=0, field_ra=0, field_dec=0):
    wcs = WCS({'CTYPE1': 'RA---TAN',
               'CTYPE2': 'DEC--TAN',
               'CRPIX1': 0,
               'CRPIX2': 0,
               'CRVAL1': field_ra,
               'CRVAL2': field_dec,
               'CUNIT1': 'deg',
               'CUNIT2': 'deg',
               'CD1_1': 1,
               'CD1_2': 0,
               'CD2_1': 0,
               'CD2_2': 1})

    here = os.path.abspath(os.path.dirname(__file__))
    ZTF_CCD_layout_fname = '%s/data/ZTF_CCD_Layout.tbl' % here
    with open(ZTF_CCD_layout_fname, 'r') as f:
        lines = f.readlines()

    is_pole = field_is_pole(field_id)

    ZTF_CCD_corners = {}
    for line in lines[1:]:
        ra_offset, dec_offset, chip = line.split()
        ra_offset, dec_offset = -float(ra_offset), float(dec_offset)
        chip = int(chip)
        ra_corner, dec_corner = wcs.all_pix2world(ra_offset, dec_offset, 1)
        if is_pole and ra_corner > 180:
            ra_corner -= 360
        if chip not in ZTF_CCD_corners.keys():
            ZTF_CCD_corners[chip] = [[ra_corner, dec_corner]]
        else:
            ZTF_CCD_corners[chip].append([ra_corner, dec_corner])

    for i in range(1, 17):
        ZTF_CCD_corners[i] = np.array(ZTF_CCD_corners[i])

    return ZTF_CCD_corners


def _calculate_two_point_midpoint(ra0, dec0, ra1, dec1, field_id=None):
    coord1 = SkyCoord(ra0 * u.deg, dec0 * u.deg, frame='icrs')
    coord2 = SkyCoord(ra1 * u.deg, dec1 * u.deg, frame='icrs')
    pa = coord1.position_angle(coord2)
    sep = coord1.separation(coord2)
    midpoint = coord1.directional_offset_by(pa, sep / 2)

    ra, dec = midpoint.ra.value, midpoint.dec.value
    if field_id and field_is_pole(field_id) and ra > 180:
        ra -= 360
    return ra, dec


def return_ZTF_RCID_corners(field_id, field_ra, field_dec):
    ZTF_CCD_corners = load_ZTF_CCD_corners(field_id, field_ra, field_dec)

    ZTF_RCID_corners = {}
    for CCD, CCD_corners_single in ZTF_CCD_corners.items():
        CCD_bottom_right = CCD_corners_single[0]
        CCD_top_right = CCD_corners_single[1]
        CCD_top_left = CCD_corners_single[2]
        CCD_bottom_left = CCD_corners_single[3]

        right_midpoint = _calculate_two_point_midpoint(*CCD_bottom_right,
                                                       *CCD_top_right,
                                                       field_id)
        left_midpoint = _calculate_two_point_midpoint(*CCD_top_left,
                                                      *CCD_bottom_left,
                                                      field_id)
        top_midpoint = _calculate_two_point_midpoint(*CCD_top_right,
                                                     *CCD_top_left,
                                                     field_id)
        bottom_midpoint = _calculate_two_point_midpoint(*CCD_bottom_right,
                                                        *CCD_bottom_left,
                                                        field_id)
        CCD_midpoint = _calculate_two_point_midpoint(*left_midpoint,
                                                     *right_midpoint,
                                                     field_id)

        for quad in np.arange(4):
            if quad == 0:
                quad_corners = np.array([CCD_top_right,
                                         top_midpoint,
                                         CCD_midpoint,
                                         right_midpoint])
            elif quad == 1:
                quad_corners = np.array([top_midpoint,
                                         CCD_top_left,
                                         left_midpoint,
                                         CCD_midpoint])
            elif quad == 2:
                quad_corners = np.array([CCD_midpoint,
                                         left_midpoint,
                                         CCD_bottom_left,
                                         bottom_midpoint])
            elif quad == 3:
                quad_corners = np.array([right_midpoint,
                                         CCD_midpoint,
                                         bottom_midpoint,
                                         CCD_bottom_right])

            rcid = 4 * (CCD - 1) + quad
            ZTF_RCID_corners[rcid] = quad_corners

    return ZTF_RCID_corners


def load_ZTF_fields():
    here = os.path.abspath(os.path.dirname(__file__))
    ZTF_fields_fname = '%s/data/ZTF_Fields.txt' % here
    ZTF_fields = np.genfromtxt(ZTF_fields_fname, skip_header=1,
                               dtype=[('id', int), ('ra', float),
                                      ('dec', float),
                                      ('ebv', float), ('galLong', float),
                                      ('galLat', float),
                                      ('eclLong', float), ('eclLat', float),
                                      ('entry', int)])
    return ZTF_fields


def test_within_CCD_corners(ra, dec, ZTF_CCD_corners):
    lower_right = ZTF_CCD_corners[1][0]
    lower_left = ZTF_CCD_corners[4][3]
    upper_right = ZTF_CCD_corners[13][1]
    upper_left = ZTF_CCD_corners[16][2]

    polygon = Polygon([lower_right,
                       lower_left,
                       upper_left,
                       upper_right])
    point = Point(ra, dec)
    return polygon.contains(point)


def return_fields(ra, dec):
    fields = []
    ZTF_fields = load_ZTF_fields()
    for ZTF_field in ZTF_fields:
        field_id, field_ra, field_dec = ZTF_field['id'], \
                                        ZTF_field['ra'], \
                                        ZTF_field['dec']
        ZTF_CCD_corners = load_ZTF_CCD_corners(field_id, field_ra, field_dec)
        if test_within_CCD_corners(ra, dec, ZTF_CCD_corners):
            fields.append(ZTF_field['id'])

    return fields


def test_within_RCID_corners(ra, dec, ZTF_RCID_corners_single):
    lower_right = ZTF_RCID_corners_single[3]
    lower_left = ZTF_RCID_corners_single[2]
    upper_right = ZTF_RCID_corners_single[0]
    upper_left = ZTF_RCID_corners_single[1]

    polygon = Polygon([lower_right,
                       lower_left,
                       upper_left,
                       upper_right])
    point = Point(ra, dec)
    return polygon.contains(point)


def return_rcid(field_id, ra, dec):
    ZTF_fields = load_ZTF_fields()
    ZTF_field = ZTF_fields[ZTF_fields['id'] == field]
    field_ra, field_dec = ZTF_field['ra'][0], ZTF_field['dec'][0]
    ZTF_RCID_corners = return_ZTF_RCID_corners(field_id, field_ra, field_dec)

    for rcid in range(64):
        ZTF_RCID_corners_single = ZTF_RCID_corners[rcid]
        if test_within_RCID_corners(ra, dec, ZTF_RCID_corners_single):
            return rcid

    print('No matching rcid found')
    return None



#! /usr/bin/env python
"""
initialize.py
Initialization of object, object_map and radec_map files.
"""


import os
import pickle
from collections import defaultdict
import numpy as np
from scipy.spatial import cKDTree
from zort.radec import lightcurve_file_is_pole


def generate_objects_file(lightcurve_file):
    objects_file = lightcurve_file.replace('.txt', '.objects')
    if os.path.exists(objects_file):
        print('%s already exists. Skipping.' % objects_file)
        return

    f_in = open(lightcurve_file, 'r')
    object_map = {}
    object_keys = ['id', 'nepochs', 'filterid',
                   'fieldid', 'rcid', 'ra', 'dec']
    with open(objects_file, 'w') as f_out:
        f_out.write('%s\n' % ','.join(object_keys))

        while True:
            line = f_in.readline()

            # Check for end of the file
            if not line:
                break

            # Only want to look at object lines
            if not line.startswith('#'):
                continue

            data = line.replace('\n', '').split()[1:]
            data_str = ','.join(data)
            f_out.write('%s\n' % data_str)

            object_id = int(data_str.split(',')[0])
            lightcurve_position = f_in.tell() - len(line)
            object_map[object_id] = lightcurve_position

    f_in.close()

    objects_file = lightcurve_file.replace('.txt', '.objects_map')
    with open(objects_file, 'wb') as fileObj:
        pickle.dump(object_map, fileObj)


def save_radec_map(DR1_object_file, radec_map):
    radec_map_filename = DR1_object_file.replace('.objects', '.radec_map')
    with open(radec_map_filename, 'wb') as fileObj:
        pickle.dump(radec_map, fileObj)


def return_radec_map_size(radec_map):
    radec_map_size = 0
    for _, v in radec_map.items():
        radec_map_size += len(v)
    return radec_map_size


def return_radec_map_filesize(radec_map):
    radec_map_filesize = 0
    for k, _ in radec_map.items():
        for _, t in radec_map[k].items():
            radec_map_filesize += t[1] - t[0]
    return radec_map_filesize


def generate_radec_map(lightcurve_file):
    radec_map_file = lightcurve_file.replace('.txt', '.radec_map')
    if os.path.exists(radec_map_file):
        print('%s already exists. Skipping.' % radec_map_file)
        return

    objects_file = lightcurve_file.replace('.txt', '.objects')

    f_in = open(objects_file, 'r')
    _ = f_in.readline()  # skip past the header

    rcid, rcid_current = None, None
    filterid, filterid_current = None, None
    ra_arr, dec_arr, object_id_arr = [], [], []
    radec_map = defaultdict(dict)

    is_pole = lightcurve_file_is_pole(lightcurve_file)

    for line in f_in:
        data = line.replace('\n', '').split(',')

        object_id = int(data[0])
        rcid = int(data[4])
        filterid = int(data[2])

        # Initialize the rcid
        if rcid_current is None:
            rcid_current = rcid
            filterid_current = filterid

        # Check to see if the block has switched
        if rcid != rcid_current:
            objects = np.array([ra_arr, dec_arr]).T
            kdtree = cKDTree(objects)
            radec_map[filterid_current][rcid_current] = \
                (kdtree, object_id_arr)

            rcid_current = rcid
            filterid_current = filterid
            ra_arr, dec_arr, object_id_arr = [], [], []

        ra, dec = float(data[5]), float(data[6])
        if is_pole and ra > 180:
            ra -= 360

        ra_arr.append(ra)
        dec_arr.append(dec)
        object_id_arr.append(object_id)

    objects = np.array([ra_arr, dec_arr]).T
    kdtree = cKDTree(objects)
    radec_map[filterid_current][rcid_current] = \
        (kdtree, object_id_arr)

    f_in.close()

    save_radec_map(radec_map_file, radec_map)



#! /usr/bin/env python
"""
object.py
Each ZTF object can be represented as an instance of the Object class, along
with its metadata and lightcurve. Note that each ZTF object is only one color,
with a different color of the same astrophysical object labelled as a different
object. This class can find and save spatially coincident objects with the
locate_siblings function.
"""
import os
import pickle
import numpy as np
from astropy.coordinates import SkyCoord
import astropy.units as u

from zort.lightcurve import Lightcurve
from zort.utils import return_filename, return_objects_filename, \
    return_objects_map_filename, return_radec_map_filename, \
    filterid_dict
from zort.plot import plot_object, plot_objects
from zort.radec import field_is_pole


################################
#                              #
#  Object Class                #
#                              #
################################

class Object:
    """
    Each ZTF object can be represented as an instance of the Object class,
    along with its parameters and lightcurve. Note that each ZTF object is only
    one color, with a different color from the same astrophysical source
    labelled as a different object. This class can find and save spatially
    coincident objects with the locate_siblings function.
    """

    def __init__(self, filename, object_id=None, lightcurve_position=None,
                 apply_catmask=False, PS_g_minus_r=0,
                 objects_map=None, radec_map=None):
        # Check object_id and lightcurve_position
        if object_id is None and lightcurve_position is None:
            raise Exception('ObjectID or lightcurve_position must be defined.')
        elif object_id is not None and lightcurve_position is not None:
            raise Exception('Only initialize object with Object ID '
                            'or lightcurve position, but not both.')

        # Load filenames and check for existence
        self.filename = return_filename(filename)
        self.objects_filename = return_objects_filename(filename)
        self.objects_map_filename = return_objects_map_filename(filename)
        self.radec_map_filename = return_radec_map_filename(filename)

        self.objects_map = None
        if object_id is not None:
            if objects_map:
                self.objects_map = objects_map
            else:
                self.objects_map = self.load_objects_map()
            self.lightcurve_position = self.objects_map[object_id]
        elif lightcurve_position is not None:
            self.lightcurve_position = lightcurve_position

        params = self._load_params()
        self.object_id = params['object_id']
        self.nepochs = params['nepochs']
        self.filterid = params['filterid']
        self.fieldid = params['fieldid']
        self.rcid = params['rcid']
        self.ra = params['ra']
        self.dec = params['dec']
        self.color = self._return_filterid_color()
        self.apply_catmask = apply_catmask
        self.PS_g_minus_r = PS_g_minus_r
        self._lightcurve = None
        self.siblings = None
        self.nsiblings = 0
        self._glonlat = None

        if radec_map:
            self.radec_map = radec_map
        else:
            self.radec_map = None

    def __repr__(self):
        title = 'Filename: %s\n' % self.filename.split('/')[-1]
        title += 'Object ID: %i\n' % self.object_id
        title += 'Filter ID: %i | Color: %s\n' % (self.filterid, self.color)
        title += 'Ra/Dec: (%.5f, %.5f)\n' % (self.ra, self.dec)
        if self.apply_catmask:
            title += '%i Epochs passing catmask\n' % self.lightcurve.nepochs
        else:
            title += '%i Epochs without applying catmask\n' % self.lightcurve.nepochs
        title += '%i siblings identified\n' % self.nsiblings
        title += 'Lightcurve Buffer Position: %i\n' % self.lightcurve_position

        return title

    @property
    def glonlat(self):
        if self._glonlat is None:
            coord = SkyCoord(self.ra, self.dec, unit=u.degree, frame='icrs')
            glon, glat = coord.galactic.l.value, coord.galactic.b.value
            if glon > 180:
                glon -= 360
            self._glonlat = (glon, glat)
        return self._glonlat

    @property
    def glon(self):
        return self.glonlat[0]

    @property
    def glat(self):
        return self.glonlat[1]

    def _load_params(self):
        # Open lightcurve file
        file = open(self.filename, 'r')

        # Jump to the location of the object in the lightcurve file
        file.seek(self.lightcurve_position)

        line = file.readline()
        params = line.split()[1:]

        params_dict = dict()
        params_dict['object_id'] = int(params[0])
        params_dict['nepochs'] = int(params[1])
        params_dict['filterid'] = int(params[2])
        params_dict['fieldid'] = int(params[3])
        params_dict['rcid'] = int(params[4])
        params_dict['ra'] = float(params[5])
        params_dict['dec'] = float(params[6])
        return params_dict

    def _return_filterid_color(self):
        # Defined by ZTF convention
        return filterid_dict[self.filterid]

    def load_objects_map(self):
        objects_map_filename = self.objects_map_filename
        objects_map = pickle.load(open(objects_map_filename, 'rb'))
        return objects_map

    def load_radec_map(self):
        radec_map_filename = self.radec_map_filename
        radec_map = pickle.load(open(radec_map_filename, 'rb'))
        return radec_map

    def return_siblings_filename(self):
        siblings_filename = self.filename.replace('.txt', '.siblings')
        return siblings_filename

    @property
    def lightcurve(self):
        if not self._lightcurve:
            self._lightcurve = self._load_lightcurve()
        return self._lightcurve

    def _load_lightcurve(self):
        return Lightcurve(self.filename, self.lightcurve_position,
                          apply_catmask=self.apply_catmask,
                          PS_g_minus_r=self.PS_g_minus_r)

    def set_siblings(self, sibling_object_ids, printFlag=False):
        # Assign the sibling to its own object instance
        siblings = []
        for object_id in sibling_object_ids:
            sib = Object(self.filename, object_id,
                         objects_map=self.objects_map,
                         radec_map=self.radec_map)
            siblings.append(sib)
            if printFlag:
                print('---- Sibling found at %.5f, %.5f !' % (
                    sib.ra, sib.dec))
                print('---- Original Color: %s | Sibling Color: %s' % (
                    self.color, sib.color))
        self.siblings = siblings
        self.nsiblings = len(siblings)

    def locate_siblings(self, radius_as=2,
                        skip_filterids=None, printFlag=False):
        radius_deg = radius_as / 3600.
        if printFlag:
            print('Locating siblings for ZTF Object %i' % self.object_id)
            print('-- Object location: %.5f, %.5f ...' % (self.ra, self.dec))

        if self.radec_map is None:
            self.radec_map = self.load_radec_map()

        # Searching for siblings in the opposite
        # filtered sections of the radec_map
        sibling_filterids = [i for i in [1, 2, 3] if i != self.filterid]
        if skip_filterids:
            sibling_filterids = [i for i in sibling_filterids
                                 if i not in skip_filterids]
        rcid = self.rcid
        siblings_object_ids = []
        is_pole = field_is_pole(self.fieldid)

        for filterid in sibling_filterids:
            color = filterid_dict[filterid]
            if filterid not in self.radec_map:
                if printFlag:
                    print('-- radec_map does not contain filter %s' % color)
                continue
            elif rcid not in self.radec_map[filterid]:
                if printFlag:
                    print('-- radec_map %s does not have rcid %i' % (color, rcid))
                continue

            kdtree, object_id_arr = self.radec_map[filterid][rcid]
            query_ra, query_dec = self.ra, self.dec
            if is_pole and query_ra > 180:
                query_ra -= 360
            idx = kdtree.query_ball_point((query_ra, query_dec), radius_deg)
            if len(idx) == 0:
                continue
            sibling_object_id = int(object_id_arr[idx[0]])

            if sibling_object_id is None:
                if printFlag:
                    print('---- No siblings found for filter %i' % filterid)
            else:
                siblings_object_ids.append(sibling_object_id)

        self.set_siblings(siblings_object_ids, printFlag)

    def plot_lightcurve(self, filename=None, insert_radius=30):
        if filename is None:
            filename = '%s_%i_lc.png' % (self.filename.replace('.txt', ''),
                                         self.object_id)

        plot_object(filename=filename, object=self,
                    insert_radius=insert_radius)

    def plot_lightcurves(self, filename=None, insert_radius=30):
        if self.siblings is None:
            self.locate_siblings()

        if filename is None:
            filename = '%s_%s_lc_with_siblings.png' % (
                self.filename.replace('.txt', ''), self.object_id)

        source_dict = {'g': None, 'r': None, 'i': None,
                       self.color: self}
        for sibling in self.siblings:
            source_dict[sibling.color] = sibling

        plot_objects(filename=filename,
                     object_g=source_dict['g'],
                     object_r=source_dict['r'],
                     object_i=source_dict['i'],
                     insert_radius=insert_radius)


def save_objects(filename, objects, overwrite=False):
    if os.path.exists(filename) and not overwrite:
        print('%s already exists, exiting without saving objects. '
              'Set overwrite=True to enable writing over existing '
              'object lists.' % filename)
        return None

    with open(filename, 'w') as f:
        for obj in objects:
            f.write('%s,%i\n' % (obj.filename, obj.lightcurve_position))


def load_objects(filename):
    objects = []
    for line in open(filename, 'r'):
        filename, lightcurve_position = line.replace('\n', '').split(',')
        objects.append(Object(filename, lightcurve_position))

    return objects



#! /usr/bin/env python
"""
utils.py
"""

import os

filterid_dict = {
    1: 'g',
    2: 'r',
    3: 'i'
}


def return_filename(filename):
    try:
        filename = filename.decode()
    except AttributeError:
        pass

    if filename is None:
        raise FileNotFoundError(filename)

    return filename


def return_objects_filename(filename):
    return _return_util_filename(filename, 'objects')


def return_objects_map_filename(filename):
    return _return_util_filename(filename, 'objects_map')


def return_radec_map_filename(filename):
    return _return_util_filename(filename, 'radec_map')


def _return_util_filename(filename, extension):
    util_filename = filename.replace('txt', extension)
    if os.path.exists(util_filename):
        return util_filename
    else:
        print('******')
        print('%s file missing, must be generated using '
              'zort-initialize to continue.' % extension)
        print("Run 'zort-initialize -h' or visit "
              "https://github.com/MichaelMedford/zort for more details.")
        print('******')
        raise FileNotFoundError(util_filename)



#! /usr/bin/env python
"""
photometry.py
A set of functions for performing photometric calculations on lightcurves.
"""

import numpy as np


def magnitudes_to_fluxes(m, sig_m, zero_point=22.0):
    """
    Given the mean and the standard deviation of a magnitude,
    assumed to be normally distributed, and a reference magnitude m0, this
    function returns the mean and the standard deviation of a Flux,
    which is log-normally distributed.
    Created by: Nathan Golovich <golovich1@llnl.gov>
    """

    # Calculate the mean and std. deviation for lnF which is assumed to be
    # normally distributed
    e = np.exp(1)
    mu_lnF = (zero_point - m) / (2.5 * np.log10(e))
    sig_lnF = sig_m / (2.5 * np.log10(e))

    # If lnF is normally distributed, F is log-normaly distributed with a mean
    # and root-variance given by
    mu_F = np.exp(mu_lnF + 0.5 * sig_lnF ** 2)
    sig_F = np.sqrt(
        (np.exp(sig_lnF ** 2) - 1) * np.exp(2 * mu_lnF + sig_lnF ** 2))

    return mu_F, sig_F


def fluxes_to_magnitudes(F, sig_F, zero_point=22.0):
    """
    Does the same thing as `magnitudes_to_fluxes` except in reverse.
    Created by: Nathan Golovich <golovich1@llnl.gov>
    """
    e = np.exp(1)
    sig_m = 2.5 * np.log10(e) * np.sqrt(np.log(sig_F ** 2 / F ** 2 + 1))
    mu_m = zero_point - 2.5 * np.log10(e) * (np.log(F) - 0.5 * np.log(
        1 + sig_F ** 2 / F ** 2))

    return mu_m, sig_m



#! /usr/bin/env python
"""
source.py
A source contains one or more Objects that are spatially coincident, each
containing a single color lightcurve of a ZTF object.
"""
import os
import pickle
import numpy as np
from astropy.coordinates import SkyCoord
import astropy.units as u

from zort.lightcurve import Lightcurve
from zort.object import Object
from zort.plot import plot_objects
from zort.utils import return_filename, return_objects_map_filename, \
    return_radec_map_filename, filterid_dict


################################
#                              #
#  Source Class                #
#                              #
################################

class Source:
    """
    A source contains one or more Objects that are spatially coincident, each
    containing a single color lightcurve of a ZTF object. All objects within
    source must be from the same lightcurve file.d
    """

    def __init__(self, filename, object_id_g=None,
                 object_id_r=None, object_id_i=None,
                 lightcurve_position_g=None,
                 lightcurve_position_r=None,
                 lightcurve_position_i=None,
                 apply_catmask=False, PS_g_minus_r=0,
                 objects_map=None,
                 radec_map=None):
        # Load filenames and check for existence
        self.filename = return_filename(filename)
        self.objects_map_filename = return_objects_map_filename(filename)
        self.radec_map_filename = return_radec_map_filename(filename)

        self._check_initialization(object_id_g, object_id_r, object_id_i,
                                   lightcurve_position_g,
                                   lightcurve_position_r,
                                   lightcurve_position_i)

        self.apply_catmask = apply_catmask
        self.PS_g_minus_r = PS_g_minus_r
        if objects_map:
            self.objects_map = objects_map
        else:
            self.objects_map = None
        if radec_map:
            self.radec_map = radec_map
        else:
            self.radec_map = None

        objects = self._load_objects(object_id_g, object_id_r, object_id_i,
                                     lightcurve_position_g,
                                     lightcurve_position_r,
                                     lightcurve_position_i)
        self.object_g = objects[0]
        self.object_r = objects[1]
        self.object_i = objects[2]
        self.objects = [o for o in objects if o is not None]

        radec = self._calculate_radec()
        self.ra = radec[0]
        self.dec = radec[1]

        self._glonlat = None

    @property
    def glonlat(self):
        if self._glonlat is None:
            coord = SkyCoord(self.ra, self.dec, unit=u.degree, frame='icrs')
            glon, glat = coord.galactic.l.value, coord.galactic.b.value
            if glon > 180:
                glon -= 360
            self._glonlat = (glon, glat)
        return self._glonlat

    @property
    def glon(self):
        return self.glonlat[0]

    @property
    def glat(self):
        return self.glonlat[1]

    def _return_object_print_info(self, object):
        if object is None:
            object_id, nepochs = None, None
        else:
            object_id, nepochs = object.object_id, object.nepochs
        return {'id': object_id, 'nepochs': nepochs}

    def __repr__(self):
        title = 'Filename: %s\n' % self.filename.split('/')[-1]
        title += 'Object-g ID: {id} | nepochs: {nepochs}\n'.format(
            **self._return_object_print_info(self.object_g))
        title += 'Object-r ID: {id} | nepochs: {nepochs}\n'.format(
            **self._return_object_print_info(self.object_r))
        title += 'Object-i ID: {id} | nepochs: {nepochs}\n'.format(
            **self._return_object_print_info(self.object_i))
        title += 'Ra/Dec: (%.5f, %.5f)\n' % (self.ra, self.dec)

        return title

    def _check_initialization(self,
                              object_id_g, object_id_r, object_id_i,
                              lightcurve_position_g,
                              lightcurve_position_r,
                              lightcurve_position_i):
        arr = np.array([object_id_g, object_id_r, object_id_i,
                        lightcurve_position_g, lightcurve_position_r,
                        lightcurve_position_i])
        if np.all(arr == None):
            raise Exception('Source must be instantiated '
                            'with at least one Object ID '
                            'or lightcurve position.')

        if object_id_g is not None and lightcurve_position_g is not None:
            raise Exception('Only initialize g object with Object ID '
                            'or lightcurve position, but not both.')

        if object_id_r is not None and lightcurve_position_r is not None:
            raise Exception('Only initialize r object with Object ID '
                            'or lightcurve position, but not both.')

        if object_id_i is not None and lightcurve_position_i is not None:
            raise Exception('Only initialize i object with Object ID '
                            'or lightcurve position, but not both.')

    def _load_object(self, object_id, lightcurve_position, color):
        if object_id is None and lightcurve_position is None:
            return None

        if lightcurve_position is not None:
            obj = Object(self.filename,
                         lightcurve_position=lightcurve_position,
                         objects_map=self.objects_map)
        else:
            if self.objects_map is None:
                self.objects_map = self.load_objects_map()
            obj = Object(self.filename,
                         object_id=object_id,
                         objects_map=self.objects_map)
        if obj.color != color:
            str = "Color of 'object_id_{color}' is {color_obj}. " \
                  "Must be {color}.".format(color=color,
                                            color_obj=obj.color)
            raise Exception(str)
        return obj

    def _load_objects(self, object_id_g, object_id_r, object_id_i,
                      lightcurve_position_g, lightcurve_position_r, lightcurve_position_i):
        object_g = self._load_object(object_id=object_id_g,
                                     lightcurve_position=lightcurve_position_g,
                                     color='g')
        object_r = self._load_object(object_id=object_id_r,
                                     lightcurve_position=lightcurve_position_r,
                                     color='r')
        object_i = self._load_object(object_id=object_id_i,
                                     lightcurve_position=lightcurve_position_i,
                                     color='i')
        return object_g, object_r, object_i

    def _calculate_radec(self):
        ra = np.mean([obj.ra for obj in self.objects if obj is not None])
        dec = np.mean([obj.dec for obj in self.objects if obj is not None])
        return ra, dec

    def load_objects_map(self):
        objects_map_filename = self.objects_map_filename
        objects_map = pickle.load(open(objects_map_filename, 'rb'))
        return objects_map

    def load_radec_map(self):
        radec_map_filename = self.radec_map_filename
        radec_map = pickle.load(open(radec_map_filename, 'rb'))
        return radec_map

    def plot_lightcurves(self, filename=None, insert_radius=30):
        if filename is None:
            object_ids = '_'.join([str(obj.object_id) for obj in self.objects
                                   if obj is not None])
            filename = '%s_%s.png' % (self.filename.replace('.txt', ''),
                                      object_ids)

        plot_objects(filename=filename,
                     object_g=self.object_g,
                     object_r=self.object_r,
                     object_i=self.object_i,
                     insert_radius=insert_radius)


def create_source_from_object(object, locate_siblings=True, radius_as=2, skip_filterids=None):
    if locate_siblings and object.siblings is None:
        object.locate_siblings(radius_as=radius_as,
                               skip_filterids=skip_filterids)
    source_dict = {'g': None, 'r': None, 'i': None,
                   object.color: object.object_id}
    for sibling in object.siblings:
        source_dict[sibling.color] = sibling.object_id

    return Source(filename=object.filename,
                  object_id_g=source_dict['g'],
                  object_id_r=source_dict['r'],
                  object_id_i=source_dict['i'])


def save_sources(filename, sources, overwrite=False):
    if os.path.exists(filename) and not overwrite:
        print('%s already exists, exiting without saving objects. '
              'Set overwrite=True to enable writing over existing '
              'object lists.' % filename)
        return None

    with open(filename, 'w') as f:
        for source in sources:
            f.write('%s,%s,%s,%s\n' % (source.filename,
                                       source.object_id_g,
                                       source.object_id_r,
                                       source.object_id_i))


def load_sources(filename):
    sources = []
    for line in open(filename, 'r'):
        filename, object_id_g, object_id_r, object_id_i = line.replace('\n', '').split(',')
        sources.append(Source(filename,
                              object_id_g=object_id_g,
                              object_id_r=object_id_r,
                              object_id_i=object_id_i))

    return sources



#! /usr/bin/env python
"""
lightcurve.py
The observations of a ZTF object make up its lightcurve, held in the
Lightcurve class. Typically this class will be instantiated as an attribute of
an Object class.
"""

import numpy as np
import os
from zort.photometry import magnitudes_to_fluxes
from zort.utils import return_filename


################################
#                              #
#  Lightcurve Class            #
#                              #
################################

class Lightcurve:
    """
    The observations of a ZTF object make up its lightcurve, held in the
    Lightcurve class. Typically this class will be instantiated as an attribute
    of an Object class.
    """

    def __init__(self, filename, lightcurve_position,
                 apply_catmask=True, PS_g_minus_r=0):
        # Load filenames and check for existence
        self.filename = return_filename(filename)

        self.lightcurve_position = lightcurve_position
        self.object_id = None  # set in self._load_lightcurve
        self.apply_catmask = apply_catmask
        self.PS_g_minus_r = PS_g_minus_r
        data = self._load_lightcurve()
        self.hmjd = data['hmjd']
        self.mag = data['mag']
        self.magerr = data['magerr']
        flux, fluxerr = magnitudes_to_fluxes(self.mag, self.magerr)
        self.flux = flux
        self.fluxerr = fluxerr
        self.clrcoeff = data['clrcoeff']
        self.catflags = data['catflags']
        self.nepochs = int(len(self.mag))
        self.mag_med = self._return_median(self.mag)
        self.mag_std = self._return_std(self.mag)
        self.flux_med = self._return_median(self.flux)
        self.flux_std = self._return_std(self.flux)

    def __repr__(self):
        title = 'Object ID: %s\n' % self.object_id
        title += 'N_epochs: %i\n' % self.nepochs

        return title

    def _load_lightcurve(self):
        """
        Loads the lightcurve from a lightcurve file, starting at the location
        of the object. The default is to apply a mask to any observations with
        a catflags >= 32768, following the recommendation of the ZTF Public Data
        Release website.
        """

        # Open file containing the lightcurve
        # Jump to the location of the object in the lightcurve file
        file = open(self.filename, 'r')
        file.seek(self.lightcurve_position)
        line = file.readline()
        self.object_id = int(line.split()[1:][0])

        data = []

        for line in file:
            if line.startswith('#'):
                break
            else:
                data.append(tuple(line.split()))

        # Assemble the data into a numpy recarray
        dtype = [('hmjd', float), ('mag', float), ('magerr', float),
                 ('clrcoeff', float), ('catflags', int)]
        data = np.array(data, dtype=dtype)

        # Apply the color correction
        data['mag'] = data['mag'] + data['clrcoeff'] * self.PS_g_minus_r

        # Apply the quality cut mask
        if self.apply_catmask:
            mask = data['catflags'] >= 32768
            data = data[~mask]

        # Sort the observations by date
        cond = np.argsort(data['hmjd'])
        data = data[cond]

        return data

    def _return_median(self, data):
        if len(data) == 0:
            return None
        else:
            return np.median(data)

    def _return_std(self, data):
        if len(data) == 0:
            return None
        else:
            return np.std(data)


def save_lightcurves(filename, lightcurves, overwrite=False):
    if os.path.exists(filename) and not overwrite:
        print('%s already exists, exiting without saving objects. '
              'Set overwrite=True to enable writing over existing '
              'object lists.' % filename)
        return None

    with open(filename, 'w') as f:
        for lightcurve in lightcurves:
            f.write('%s,%i\n' % (lightcurve.filename,
                                 lightcurve.lightcurve_position))


def load_lightcurves(filename):
    lightcurves = []
    for line in open(filename, 'r'):
        filename, lightcurve_position = line.replace('\n', '').split(',')
        lightcurves.append(Lightcurve(filename, lightcurve_position))

    return lightcurves



#! /usr/bin/env python
"""
plot.py
"""

import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt


def _plot_axis(ax, object, hmjd_min, hmjd_max, insert_radius):
    if object.color == 'i':
        color = 'k'
    else:
        color = object.color

    ax[0].errorbar(object.lightcurve.hmjd, object.lightcurve.mag,
                   yerr=object.lightcurve.magerr,
                   ls='none', marker='.', color=color)
    ax[0].invert_yaxis()
    ax[0].set_xlim(hmjd_min, hmjd_max)
    ax[0].set_ylabel('Magnitude')
    ax[0].set_xlabel('Observation Date')
    ax[0].set_title('ZTF Object %i (%s band)' % (object.object_id,
                                                 object.color))

    hmjd0 = object.lightcurve.hmjd[np.argmin(object.lightcurve.mag)]
    hmjd_min_insert = hmjd0 - insert_radius
    hmjd_max_insert = hmjd0 + insert_radius
    hmjd_cond = (object.lightcurve.hmjd >= hmjd_min_insert) & (
            object.lightcurve.hmjd <= hmjd_max_insert)

    ax[1].errorbar(object.lightcurve.hmjd[hmjd_cond],
                   object.lightcurve.mag[hmjd_cond],
                   yerr=object.lightcurve.magerr[hmjd_cond],
                   ls='none', marker='.', color=color)
    ax[1].invert_yaxis()
    ax[1].set_xlim(hmjd_min_insert, hmjd_max_insert)
    ax[1].set_ylabel('Magnitude')
    ax[1].set_xlabel('Observation Date')
    ax[1].set_title('%i Days Around Peak' % insert_radius)


def plot_object(filename, object, insert_radius=30):
    hmjd_min = np.min(object.lightcurve.hmjd) - 10
    hmjd_max = np.max(object.lightcurve.hmjd) + 10

    fig, ax = plt.subplots(1, 2, figsize=(12, 4))
    fig.subplots_adjust(hspace=0.4)

    _plot_axis(ax, object, hmjd_min, hmjd_max, insert_radius)

    fig.savefig(filename, dpi=300, bbox_inches='tight', pad_inches=0.05)
    print('---- Lightcurve saved: %s' % filename)
    plt.close(fig)


def plot_objects(filename, object_g=None, object_r=None,
                 object_i=None, insert_radius=30):
    if object_g is None and object_r is None and object_i is None:
        raise Exception('At least one object must be set to generate lightcurves.')

    objects = [object_g, object_r, object_i]
    objects = [obj for obj in objects if obj is not None]

    if len(objects) == 1:
        plot_object(objects[0], filename)
    elif len(objects) == 2:
        N_rows = 2
        figsize_height = 6
    else:
        N_rows = 3
        figsize_height = 9

    hmjd_min = min([o.lightcurve.hmjd.min() for o in objects]) - 10
    hmjd_max = max([o.lightcurve.hmjd.max() for o in objects]) + 10

    fig, ax = plt.subplots(N_rows, 2, figsize=(12, figsize_height))
    fig.subplots_adjust(top=0.92)
    fig.subplots_adjust(hspace=0.4)

    for i, object in enumerate(objects):
        _plot_axis(ax[i], object, hmjd_min, hmjd_max, insert_radius)

    fig.savefig(filename, dpi=300, bbox_inches='tight', pad_inches=0.05)
    print('---- Lightcurves saved: %s' % filename)
    plt.close(fig)



#! /usr/bin/env python
"""
__setup__.py
Inspired by: https://github.com/pypa/sampleproject
"""

DESCRIPTION = "Reader for lightcurves from the ZTF Public Data Release"
DISTNAME = 'zort'
AUTHOR = 'Michael Medford'
AUTHOR_EMAIL = 'michaelmedford@berkeley.edu'
URL = 'https://github.com/MichaelMedford/zort'
LICENSE = 'MIT'
VERSION = open('VERSION').readline().strip()
DOWNLOAD_URL = 'https://github.com/MichaelMedford/zort/tarball/%s' % VERSION

from setuptools import setup, find_packages
from os import path

# Get the long description from the README file
here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()
LONG_DESCRIPTION_CONTENT_TYPE = 'text/markdown'

setup(name=DISTNAME,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
      license=LICENSE,
      url=URL,
      version=VERSION,
      download_url=DOWNLOAD_URL,
      packages=find_packages(exclude=['contrib', 'docs', 'tests']),
      package_data={'zort': ['data/ZTF_Fields.txt',
                             'data/ZTF_CCD_Layout.tbl']},
      python_requires='>=3.5',
      install_requires=['tqdm',
                        'matplotlib',
                        'numpy',
                        'scipy',
                        'shapely'],
      scripts=['bin/zort-initialize'],
      classifiers=['Intended Audience :: Science/Research',
                   'Programming Language :: Python :: 3.6',
                   'License :: OSI Approved :: MIT License',
                   'Topic :: Scientific/Engineering :: Astronomy',
                   'Operating System :: POSIX',
                   'Operating System :: Unix',
                   'Operating System :: MacOS'],
      project_urls={'Lightcurves': 'https://www.ztf.caltech.edu/page/dr1#12c'}
      )



#! /usr/bin/env python
"""
parallel.py
A native python implementation of processing embarrassingly parallel tasks
using concurrent futures. These processes will also display progress bars.
"""
from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor, as_completed


def parallel_process(array, function, n_jobs=16, use_kwargs=False, return_out=False):
    """
        A parallel version of the map function with a progress bar.

        Args:
            array (array-like): An array to iterate over.
            function (function): A python function to apply to the elements of array
            n_jobs (int, default=16): The number of cores to use
            use_kwargs (boolean, default=False): Whether to consider the elements of array as dictionaries of
                keyword arguments to function
        Returns:
            [function(array[0]), function(array[1]), ...]
        Function inspired by:
            http://danshiebler.com/2016-09-14-parallel-progress-bar/
    """
    # Assemble the workers
    with ProcessPoolExecutor(max_workers=n_jobs) as pool:
        # Pass the elements of array into function
        if use_kwargs:
            futures = [pool.submit(function, **a) for a in array]
        else:
            futures = [pool.submit(function, a) for a in array]
        kwargs = {
            'total': len(futures),
            'unit': 'it',
            'unit_scale': True,
            'leave': True
        }
        # Print out the progress as tasks complete
        for _ in tqdm(as_completed(futures), **kwargs):
            pass

    if return_out:
        out = []
        # Get the results from the futures.
        for i, future in tqdm(enumerate(futures)):
            try:
                out.append(future.result())
            except Exception as e:
                out.append(e)
        return out



#! /usr/bin/env python
"""
lightcurveFile.py
ZTF objects are stored in lightcurve files. This class allows for
iteration through a lightcurve file in order to inspect each object's
lightcurve. Filters can be applied to objects and/or lightcurves.
"""

import os
import pickle

from zort.object import Object
from zort.utils import return_filename
from zort.utils import return_objects_filename, \
    return_radec_map_filename, \
    return_objects_map_filename
from zort.radec import return_rcid


################################
#                              #
#  LightcurveFile Class        #
#                              #
################################

class LightcurveFile:
    """
    ZTF objects are stored in lightcurve files. This class allows for
    iteration through a lightcurve file in order to inspect each object's
    lightcurve. The user can select to start looping through the file at a
    specific object position.

    LightcurveFile should be used as an iterator to loop over all of the
    objects in the lightcurve file. Here is an example for how to
    (1) filter through the lightcurves in a lightcurve file and (2) recover
    the lightcurves of objects that pass a customized filter.

    The Object class (object.py) and the Lightcurve class (lightcurve.py) list
    out the attributes that can be run through the filter. These attributes
    include the object's number of epochs, filter, and sky location. Filters
    will almost certainly want to filter on the object's lightcurve, which
    contains the observation dates, magnitudes and magnitude errors.

    EXAMPLE 1:

    filename = 'field000000_lightcurve_file_example.txt'
    interesting_objects = []

    from zort.lightcurveFile import LightcurveFile
    for obj in LightcurveFile(filename):
        if my_interesting_filter(obj):
            interesting_objects.append(obj)

    from zort.object import save_objects
    save_objects('objects.list', interesting_objects)

    from zort.object import load_objects
    interesting_objects = load_objects('objects.list')
    for obj in interesting_objects:
        print(obj)
        print(obj.lightcurve)

    EXAMPLE 2:

    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    rank = comm.rank
    size = comm.size

    filename = 'field000000_lightcurve_file_example.txt'
    interesting_objects = []

    from zort.lightcurveFile import LightcurveFile
    for obj in LightcurveFile(filename, proc_rank=rank, proc_size=size):
        if my_interesting_filter(obj):
            interesting_objects.append(obj)

    from zort.object import save_objects
    save_objects('objects.%i.list' % rank, interesting_objects)
    """

    def __init__(self, filename, init_object_position=40,
                 proc_rank=0, proc_size=1, apply_catmask=True):
        # Load filenames and check for existence
        self.filename = return_filename(filename)
        self.objects_filename = return_objects_filename(filename)
        self.objects_map_filename = return_objects_map_filename(filename)
        self.radec_map_filename = return_radec_map_filename(filename)

        self.init_object_position = init_object_position
        self.objects_file = self.return_objects_file()
        self.objects_map = self.load_objects_map()
        self.radec_map = self.load_radec_map()
        self.proc_rank = proc_rank
        self.proc_size = proc_size
        self.fieldid = int(os.path.basename(filename).split('_')[0].
                           replace('field', ''))
        self.objects_file_counter = 0
        self.apply_catmask = apply_catmask

    def __iter__(self):
        return self

    def __next__(self):
        while self.objects_file_counter % self.proc_size != self.proc_rank:
            line = self._return_objects_file_line()
            if line == '':
                raise StopIteration

        line = self._return_objects_file_line()
        if line == '':
            raise StopIteration
        return self.return_object(line)

    def __exit__(self):
        self.objects_file.close()

    def _return_objects_file_line(self):
        line = self.objects_file.readline()
        self.objects_file_counter += 1
        return line

    def _return_parsed_line(self, line):
        return line.replace('\n', '').split(',')

    def return_filename(self, filename):
        try:
            filename = filename.decode()
        except AttributeError:
            pass

        if filename is None:
            raise FileNotFoundError(filename)

        return filename

    def return_object(self, line):
        object_id = int(self._return_parsed_line(line)[0])
        return Object(self.filename,
                      object_id=object_id,
                      objects_map=self.objects_map,
                      radec_map=self.radec_map,
                      apply_catmask=self.apply_catmask)

    def return_objects_file(self):
        file = open(self.objects_filename, 'r')
        file.seek(self.init_object_position)
        return file

    def load_objects_map(self):
        objects_map = pickle.load(open(self.objects_map_filename, 'rb'))
        return objects_map

    def load_radec_map(self):
        radec_map = pickle.load(open(self.radec_map_filename, 'rb'))
        return radec_map

    def locate_objects_by_radec(self, ra, dec, rcid=None, radius_as=2):
        if rcid is None:
            rcid = return_rcid(self.fieldid, ra, dec)
            if rcid is None:
                return

        radius_deg = radius_as / 3600.
        objects = []
        for filterid in self.radec_map:
            kdtree, lightcurve_position_arr = self.radec_map[filterid][rcid]
            idx_arr = kdtree.query_ball_point((ra, dec), radius_deg)
            if len(idx_arr) == 0:
                continue
            for idx in idx_arr:
                lightcurve_position = int(lightcurve_position_arr[idx])
                obj = Object(self.filename, lightcurve_position,
                             radec_map=self.radec_map, apply_catmask=self.apply_catmask)
                objects.append(obj)
        return objects

    def _extract_object_by_color(self, objects, color):
        objects_color = [obj for obj in objects if obj.color == color]
        if len(objects_color) == 0:
            return None
        elif len(objects_color) == 1:
            return objects_color[0]
        else:
            print('Multiple objects found for filter %s. '
                  'Returning first.' % color)
            return objects_color[0]

    def locate_source_by_radec(self, ra, dec, rcid=None, radius_as=2):
        objects = locate_objects_by_radec(ra, dec, rcid, radius_as)
        g_object = _extract_object_by_color(objects, 'g')
        r_object = _extract_object_by_color(objects, 'r')
        i_object = _extract_object_by_color(objects, 'i')
        if g_object is None and r_object is None and i_object is None:
            return None
        source = Source(object_id_g=g_object.id,
                        object_id_r=r_object.id,
                        object_id_i=i_object.id,
                        apply_catmask=self.apply_catmask,
                        radec_map=self.radec_map)
        return source
